/* This file has been placed in the public domain. */

/* required to make the slide show run at all */
@import url(s5-core.css);

/* sets basic placement and size of slide components */
@import url(framing.css);

/* styles that make the slides look good */
@import url(pretty.css);
